from __future__ import annotations

from typing import Any

from ikea_api.abc import Endpoint, SessionInfo, endpoint
from ikea_api.base_ikea_api import BaseIkeaAPI
from ikea_api.error_handlers import handle_json_decode_error
from ikea_api.exceptions import ItemFetchError


def build_url(item_code: str, is_combination: bool) -> str:
    prefix = "s" if is_combination else ""
    return f"/{item_code[5:]}/{prefix}{item_code}.json"


class PipItem(BaseIkeaAPI):
    def _get_session_info(self) -> SessionInfo:
        url = f"{self._const.local_base_url}/products"
        headers = self._extend_default_headers({"Accept": "*/*"})
        return SessionInfo(base_url=url, headers=headers)

    @endpoint()
    def get_item(
        self, item_code: str, is_combination: bool = True
    ) -> Endpoint[dict[str, Any]]:
        response = yield self._RequestInfo("GET", build_url(item_code, is_combination))

        if response.status_code == 404 and is_combination:
            response = yield self._RequestInfo("GET", build_url(item_code, False))
            if response.status_code == 404:
                raise ItemFetchError(response)

        handle_json_decode_error(response)
        return response.json



from __future__ import annotations

import datetime
from typing import List, Optional

from pydantic import BaseModel, HttpUrl


class ChildItem(BaseModel):
    name: Optional[str]
    item_code: str
    weight: float
    qty: int


class ParsedItem(BaseModel):
    is_combination: bool
    item_code: str
    name: str
    image_url: Optional[str]
    weight: float
    child_items: List[ChildItem]
    price: int
    url: str
    category_name: Optional[str]
    category_url: Optional[HttpUrl]


class IngkaItem(BaseModel):
    is_combination: bool
    item_code: str
    name: str
    image_url: Optional[str]
    weight: float
    child_items: List[ChildItem]


class PipItem(BaseModel):
    item_code: str
    price: int
    url: str
    category_name: Optional[str]
    category_url: Optional[HttpUrl]


class UnavailableItem(BaseModel):
    item_code: str
    available_qty: int


class DeliveryService(BaseModel):
    is_available: bool
    date: Optional[datetime.date]
    type: str
    price: int
    service_provider: Optional[str]
    unavailable_items: List[UnavailableItem]


class GetDeliveryServicesResponse(BaseModel):
    delivery_options: List[DeliveryService]
    cannot_add: List[str]


class CostsOrder(BaseModel):
    delivery_cost: float
    total_cost: float


class StatusBannerOrder(BaseModel):
    purchase_date: datetime.date
    delivery_date: datetime.date


class PurchaseInfo(StatusBannerOrder, CostsOrder):
    pass


class PurchaseHistoryItem(BaseModel):
    id: str
    status: str
    price: int
    datetime: str
    datetime_formatted: str
    store: str


CannotAddItems = List[str]



from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property, lru_cache
from typing import TYPE_CHECKING, Any, cast

from ikea_api.abc import (
    AsyncExecutor,
    EndpointInfo,
    EndpointResponse,
    RequestInfo,
    ResponseInfo,
    SessionInfo,
)

if TYPE_CHECKING:
    import httpx


@dataclass
class HttpxResponseInfo(ResponseInfo):
    response: httpx.Response

    def __post_init__(self) -> None:
        self.headers = cast("httpx.Headers", self.response.headers)  # type: ignore
        self.status_code = cast(int, self.response.status_code)  # type: ignore

    @cached_property
    def text(self) -> str:
        return self.response.text

    @cached_property
    def json(self) -> Any:
        return self.response.json()

    @property
    def is_success(self) -> bool:
        return self.response.is_success


@lru_cache
def get_cached_session(headers: frozenset[tuple[str, str]]) -> httpx.AsyncClient:
    try:
        import httpx
    except ImportError:
        raise RuntimeError(
            "To use httpx executor you need httpx to be installed. "
            + "Run 'pip install \"ikea_api[httpx]\"' to do so."
        )

    return httpx.AsyncClient(headers=dict(headers))


def get_session_from_info(session_info: SessionInfo) -> httpx.AsyncClient:
    return get_cached_session(headers=frozenset(session_info.headers.items()))


class HttpxExecutor(AsyncExecutor):
    @staticmethod
    async def request(request: RequestInfo) -> HttpxResponseInfo:
        session = get_session_from_info(request.session_info)
        response = await session.request(  # type: ignore
            method=request.method,
            url=request.session_info.base_url + request.url,
            params=request.params,
            data=request.data,
            json=request.json,
            headers=request.headers,
        )
        return HttpxResponseInfo(response)


async def run_async(endpoint: EndpointInfo[EndpointResponse]) -> EndpointResponse:
    return await HttpxExecutor.run(endpoint)



from __future__ import annotations

import sys
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from functools import cached_property, partial
from typing import (
    Any,
    Callable,
    Generator,
    Generic,
    Iterable,
    Literal,
    Mapping,
    TypeVar,
)

if sys.version_info < (3, 10):
    from typing_extensions import ParamSpec
else:
    from typing import ParamSpec


@dataclass
class SessionInfo:
    base_url: str
    headers: dict[str, str]


@dataclass
class RequestInfo:
    session_info: SessionInfo
    method: Literal["GET", "POST"]
    url: str
    params: dict[str, Any]
    headers: dict[str, str]
    data: Any = None
    json: Any = None


@dataclass
class ResponseInfo(ABC):
    headers: Mapping[str, str] = field(init=False)
    status_code: int = field(init=False)

    @cached_property
    @abstractmethod
    def text(self) -> str:
        ...

    @cached_property
    @abstractmethod
    def json(self) -> Any:
        ...

    @property
    @abstractmethod
    def is_success(self) -> bool:
        ...


EndpointResponse = TypeVar("EndpointResponse")
Endpoint = Generator[RequestInfo, ResponseInfo, EndpointResponse]

ErrorHandler = Callable[[ResponseInfo], None]


@dataclass
class EndpointInfo(Generic[EndpointResponse]):
    func: partial[Endpoint[EndpointResponse]]
    handlers: Iterable[ErrorHandler]


P = ParamSpec("P")


def endpoint(
    handlers: Iterable[ErrorHandler] | None = None,
) -> Callable[
    [Callable[P, Endpoint[EndpointResponse]]],
    Callable[P, EndpointInfo[EndpointResponse]],
]:
    def decorator(
        func: Callable[P, Endpoint[EndpointResponse]]
    ) -> Callable[P, EndpointInfo[EndpointResponse]]:
        def wrapper(
            *args: P.args, **kwargs: P.kwargs
        ) -> EndpointInfo[EndpointResponse]:
            return EndpointInfo(
                func=partial(func, *args, **kwargs), handlers=handlers or ()
            )

        return wrapper

    return decorator


class SyncExecutor(ABC):
    @staticmethod
    @abstractmethod
    def request(request: RequestInfo) -> ResponseInfo:
        ...

    @classmethod
    def run(cls, endpoint: EndpointInfo[EndpointResponse]) -> EndpointResponse:
        gen = endpoint.func()
        req_info = next(gen)

        while True:
            response_info = cls.request(req_info)

            try:
                for handler in endpoint.handlers:
                    handler(response_info)
                req_info = gen.send(response_info)

            except StopIteration as exc:
                return exc.value


class AsyncExecutor(ABC):
    @staticmethod
    @abstractmethod
    async def request(request: RequestInfo) -> ResponseInfo:
        ...

    @classmethod
    async def run(cls, endpoint: EndpointInfo[EndpointResponse]) -> EndpointResponse:
        gen = endpoint.func()
        req_info = next(gen)

        while True:
            response_info = await cls.request(req_info)

            try:
                for handler in endpoint.handlers:
                    handler(response_info)
                req_info = gen.send(response_info)

            except StopIteration as exc:
                return exc.value


class BaseAPI(ABC):
    _session_info: SessionInfo

    def __init__(self) -> None:
        self._session_info = self._get_session_info()

    @abstractmethod
    def _get_session_info(self) -> SessionInfo:
        pass

    def _RequestInfo(
        self,
        method: Literal["GET", "POST"],
        url: str | None = None,
        params: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
        json: Any = None,
        data: Any = None,
    ) -> RequestInfo:
        return RequestInfo(
            session_info=self._session_info,
            method=method,
            url=url or "",
            params=params or {},
            headers=headers or {},
            data=data,
            json=json,
        )



from __future__ import annotations

from typing import Any, Dict, List, cast

from ikea_api.abc import ResponseInfo


class APIError(Exception):
    """Generic API related exception."""

    response: ResponseInfo

    def __init__(self, response: ResponseInfo, msg: Any = None) -> None:
        self.response = response
        if msg is None:
            msg = (response.status_code, response.text)
        super().__init__(msg)


class JSONError(APIError):
    pass


class AuthError(APIError):
    pass


class NotSuccessError(APIError):
    pass


class GraphQLError(APIError):
    errors: list[dict[str, Any]]

    def __init__(self, response: ResponseInfo) -> None:
        if isinstance(response.json, dict):
            self.errors = response.json["errors"]
        else:
            assert isinstance(response.json, list)
            self.errors = []

            for chunk in cast(List[Dict[str, Any]], response.json):
                if "errors" in chunk:
                    self.errors += chunk["errors"]

        super().__init__(response, self.errors)


class ItemFetchError(APIError):
    pass


class WrongItemCodeError(ItemFetchError):
    pass


class ProcessingError(APIError):
    pass


class ParsingError(Exception):
    pass



from __future__ import annotations

import asyncio
from typing import Any, Iterable, List, Optional

from pydantic import BaseModel

from ikea_api.constants import Constants
from ikea_api.endpoints.cart import Cart
from ikea_api.endpoints.ingka_items import IngkaItems
from ikea_api.endpoints.iows_items import IowsItems
from ikea_api.endpoints.order_capture import (
    OrderCapture,
    convert_cart_to_checkout_items,
)
from ikea_api.endpoints.pip_item import PipItem
from ikea_api.endpoints.purchases import Purchases
from ikea_api.exceptions import GraphQLError
from ikea_api.executors.httpx import run_async as run_with_httpx
from ikea_api.executors.requests import run as run_with_requests
from ikea_api.utils import parse_item_codes, unshorten_urls_from_ingka_pagelinks
from ikea_api.wrappers import types
from ikea_api.wrappers.parsers.ingka_items import parse_ingka_items
from ikea_api.wrappers.parsers.iows_items import parse_iows_item
from ikea_api.wrappers.parsers.order_capture import parse_delivery_services
from ikea_api.wrappers.parsers.pip_item import parse_pip_item
from ikea_api.wrappers.parsers.purchases import (
    parse_costs_order,
    parse_history,
    parse_status_banner_order,
)


def get_purchase_history(purchases: Purchases) -> list[types.PurchaseHistoryItem]:
    response = run_with_requests(purchases.history())
    return parse_history(purchases._const, response)


def get_purchase_info(
    purchases: Purchases, *, order_number: str, email: str | None = None
) -> types.PurchaseInfo:
    endpoint = purchases.order_info(
        order_number=order_number,
        email=email,
        queries=["StatusBannerOrder", "CostsOrder"],
    )
    status_banner, costs = run_with_requests(endpoint)
    return types.PurchaseInfo(
        **parse_status_banner_order(status_banner).dict(),
        **parse_costs_order(costs).dict(),
    )


class _ExtensionsData(BaseModel):
    itemNos: List[str]


class _Extensions(BaseModel):
    code: str
    data: Optional[_ExtensionsData]


class _CartErrorRef(BaseModel):
    extensions: _Extensions


def add_items_to_cart(cart: Cart, items: dict[str, int]) -> types.CannotAddItems:
    run_with_requests(cart.clear())
    cannot_add_items: list[str] = []
    pending_items = items.copy()

    while pending_items:
        try:
            run_with_requests(cart.add_items(pending_items))
            break
        except GraphQLError as exc:
            for error_dict in exc.errors:
                error = _CartErrorRef(**error_dict)
                if error.extensions.code != "INVALID_ITEM_NUMBER":
                    continue
                if not error.extensions.data:
                    continue
                cannot_add_items += error.extensions.data.itemNos

            for item_code in cannot_add_items:
                pending_items.pop(item_code, None)

    return cannot_add_items


async def get_delivery_services(
    constants: Constants,
    token: str,
    items: dict[str, int],
    zip_code: str,
) -> types.GetDeliveryServicesResponse:
    cart = Cart(constants, token=token)
    order_capture = OrderCapture(constants, token=token)

    cannot_add = add_items_to_cart(cart, items)
    cannot_add_all_items = not set(items.keys()) ^ set(cannot_add)
    if cannot_add_all_items:
        return types.GetDeliveryServicesResponse(
            delivery_options=[], cannot_add=cannot_add
        )

    cart_response = await run_with_httpx(cart.show())
    checkout_items = convert_cart_to_checkout_items(cart_response)
    checkout_id = await run_with_httpx(order_capture.get_checkout(checkout_items))
    service_area_id = await run_with_httpx(
        order_capture.get_service_area(checkout_id, zip_code=zip_code)
    )

    home, collect = await asyncio.gather(
        run_with_httpx(
            order_capture.get_home_delivery_services(checkout_id, service_area_id),
        ),
        run_with_httpx(
            order_capture.get_collect_delivery_services(checkout_id, service_area_id)
        ),
    )

    parsed_data = parse_delivery_services(
        constants=constants,
        home_response=home,
        collect_response=collect,
    )
    return types.GetDeliveryServicesResponse(
        delivery_options=parsed_data, cannot_add=cannot_add
    )


def chunks(list_: list[Any], chunk_size: int) -> Iterable[list[Any]]:
    return (list_[i : i + chunk_size] for i in range(0, len(list_), chunk_size))


async def _get_ingka_items(
    constants: Constants, item_codes: list[str]
) -> list[types.IngkaItem]:
    api = IngkaItems(constants)
    tasks = (run_with_httpx(api.get_items(c)) for c in chunks(item_codes, 50))
    responses = await asyncio.gather(*tasks)
    res: list[types.IngkaItem] = []
    for response in responses:
        res += parse_ingka_items(constants, response)
    return res


async def _get_pip_items(
    constants: Constants, item_codes: list[str]
) -> list[types.PipItem | None]:
    api = PipItem(constants)
    tasks = (run_with_httpx(api.get_item(i)) for i in item_codes)
    responses = await asyncio.gather(*tasks)
    return [parse_pip_item(r) for r in responses]


def _get_pip_items_map(items: list[types.PipItem | None]) -> dict[str, types.PipItem]:
    res: dict[str, types.PipItem] = {}
    for item in items:
        if item:
            res[item.item_code] = item
    return res


async def _get_ingka_pip_items(
    constants: Constants, item_codes: list[str]
) -> list[types.ParsedItem]:
    ingka_resp, pip_resp = await asyncio.gather(
        _get_ingka_items(constants=constants, item_codes=item_codes),
        _get_pip_items(constants=constants, item_codes=item_codes),
    )
    pip_items_map = _get_pip_items_map(pip_resp)

    res: list[types.ParsedItem] = []
    for ingka_item in ingka_resp:
        pip_item_ = pip_items_map.get(ingka_item.item_code)
        if pip_item_ is None:
            continue
        item = types.ParsedItem(
            is_combination=ingka_item.is_combination,
            item_code=ingka_item.item_code,
            name=ingka_item.name,
            image_url=ingka_item.image_url,
            weight=ingka_item.weight,
            child_items=ingka_item.child_items,
            price=pip_item_.price,
            url=pip_item_.url,
            category_name=pip_item_.category_name,
            category_url=pip_item_.category_url,
        )
        res.append(item)
    return res


async def _get_iows_items(
    constants: Constants, item_codes: list[str]
) -> list[types.ParsedItem]:
    api = IowsItems(constants)
    tasks = (run_with_httpx(api.get_items(c)) for c in chunks(item_codes, 90))
    responses = await asyncio.gather(*tasks)

    res: list[types.ParsedItem] = []
    for items in responses:
        res += (parse_iows_item(constants, i) for i in items)
    return res


async def get_items(
    constants: Constants, item_codes: list[str]
) -> list[types.ParsedItem]:
    item_codes_ = item_codes.copy()
    item_codes_ += await unshorten_urls_from_ingka_pagelinks(
        str(item_codes)
    )  # TODO: Don't do this
    pending = parse_item_codes(item_codes_)
    ingka_pip = await _get_ingka_pip_items(constants=constants, item_codes=pending)

    fetched = {i.item_code for i in ingka_pip}
    pending = [i for i in pending if i not in fetched]
    if not pending:
        return ingka_pip

    iows = await _get_iows_items(constants=constants, item_codes=pending)
    return ingka_pip + iows



from __future__ import annotations

from typing import Any, TypedDict

from ikea_api.abc import Endpoint, SessionInfo, endpoint
from ikea_api.base_ikea_api import BaseAuthIkeaAPI
from ikea_api.error_handlers import (
    handle_401,
    handle_json_decode_error,
    handle_not_success,
)
from ikea_api.exceptions import ProcessingError


class CheckoutItem(TypedDict):
    itemNo: str
    quantity: int
    uom: str


handlers = (handle_json_decode_error, handle_401, handle_not_success)


class OrderCapture(BaseAuthIkeaAPI):
    def _get_session_info(self) -> SessionInfo:
        host = "ikea.ru" if self._const.country == "ru" else "ingka.com"
        url = f"https://ordercapture.{host}/ordercaptureapi/{self._const.country}"
        headers = self._extend_default_headers_with_auth(
            {"X-Client-Id": "af2525c3-1779-49be-8d7d-adf32cac1934"}
        )
        return SessionInfo(base_url=url, headers=headers)

    @endpoint(handlers)
    def get_checkout(self, items: list[CheckoutItem]) -> Endpoint[str]:
        """Generate checkout for items"""
        response = yield self._RequestInfo(
            "POST",
            "/checkouts",
            headers={"X-Client-Id": "6a38e438-0bbb-4d4f-bc55-eb314c2e8e23"},
            json={
                "channel": "WEBAPP",
                "checkoutType": "STANDARD",
                "shoppingType": "ONLINE",
                "deliveryArea": None,
                "items": items,
                "languageCode": self._const.language,
                "serviceArea": None,
                "preliminaryAddressInfo": None,
            },
        )
        if "resourceId" not in response.json:
            raise ProcessingError(response, "No id for service area")
        return response.json["resourceId"]

    @endpoint(handlers)
    def get_service_area(
        self, checkout_id: str, zip_code: str, state_code: str | None = None
    ) -> Endpoint[str]:
        """Generate delivery area for checkout from Zip Code and State Code"""
        payload = {"zipCode": zip_code}
        if state_code:
            payload["stateCode"] = state_code
        response = yield self._RequestInfo(
            "POST", f"/checkouts/{checkout_id}/service-area", json=payload
        )

        if "id" not in response.json:
            raise ProcessingError(response, "No id for service area")
        return response.json["id"]

    @endpoint(handlers)
    def get_home_delivery_services(
        self, checkout_id: str, service_area_id: str
    ) -> Endpoint[dict[str, Any]]:
        """Get available home delivery services"""
        response = yield self._RequestInfo(
            "GET",
            f"/checkouts/{checkout_id}/service-area/{service_area_id}/home-delivery-services",
        )
        return response.json

    @endpoint(handlers)
    def get_collect_delivery_services(
        self, checkout_id: str, service_area_id: str
    ) -> Endpoint[dict[str, Any]]:
        """Get available collect delivery services"""
        response = yield self._RequestInfo(
            "GET",
            f"/checkouts/{checkout_id}/service-area/{service_area_id}/collect-delivery-services",
        )
        return response.json


def convert_cart_to_checkout_items(cart_response: dict[str, Any]) -> list[CheckoutItem]:
    if not cart_response["data"].get("cart", {}).get("items"):
        raise RuntimeError("Cannot get items for Order Capture")
    return [
        {
            "quantity": item["quantity"],
            "itemNo": item["itemNo"],
            "uom": item["product"]["unitCode"],
        }
        for item in cart_response["data"]["cart"]["items"]
    ]



from __future__ import annotations

from typing import Any

from ikea_api.abc import Endpoint, SessionInfo, endpoint
from ikea_api.base_ikea_api import BaseIkeaAPI
from ikea_api.error_handlers import (
    handle_401,
    handle_json_decode_error,
    handle_not_success,
)
from ikea_api.exceptions import ItemFetchError


class IngkaItems(BaseIkeaAPI):
    def _get_session_info(self):
        headers = self._extend_default_headers(
            {
                "Accept": "*/*",
                "Referer": f"{self._const.local_base_url}/order/delivery/",
                "X-Client-Id": "c4faceb6-0598-44a2-bae4-2c02f4019d06",
            }
        )
        url = f"https://api.ingka.ikea.com/salesitem/communications/{self._const.country}/{self._const.language}"
        return SessionInfo(base_url=url, headers=headers)

    @endpoint(handlers=[handle_json_decode_error, handle_401, handle_not_success])
    def get_items(self, item_codes: list[str]) -> Endpoint[dict[str, Any]]:
        response = yield self._RequestInfo("GET", params={"itemNos": item_codes})

        if "error" in response.json:
            try:
                msg = response.json["error"]["details"][0]["value"]["keys"]
            except (KeyError, TypeError, IndexError):
                msg = None
            raise ItemFetchError(response, msg)

        return response.json



from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property, lru_cache
from typing import TYPE_CHECKING, Any

from ikea_api.abc import (
    EndpointInfo,
    EndpointResponse,
    RequestInfo,
    ResponseInfo,
    SessionInfo,
    SyncExecutor,
)

if TYPE_CHECKING:
    import requests


@dataclass
class RequestsResponseInfo(ResponseInfo):
    response: requests.Response

    def __post_init__(self) -> None:
        self.headers = self.response.headers
        self.status_code = self.response.status_code

    @cached_property
    def text(self) -> str:
        return self.response.text

    @cached_property
    def json(self) -> Any:
        return self.response.json()

    @property
    def is_success(self) -> bool:
        return self.response.ok


@lru_cache
def get_cached_session(headers: frozenset[tuple[str, str]]) -> requests.Session:
    try:
        import requests
    except ImportError:
        raise RuntimeError(
            "To use requests executor you need requests to be installed. "
            + "Run 'pip install \"ikea_api[requests]\"' to do so."
        )

    session = requests.Session()
    session.headers.update(headers)
    return session


def get_session_from_info(session_info: SessionInfo) -> requests.Session:
    return get_cached_session(headers=frozenset(session_info.headers.items()))


class RequestsExecutor(SyncExecutor):
    @staticmethod
    def request(request: RequestInfo) -> RequestsResponseInfo:
        session = get_session_from_info(request.session_info)
        response = session.request(
            method=request.method,
            url=request.session_info.base_url + request.url,
            params=request.params,
            data=request.data,
            json=request.json,
            headers=request.headers,
        )
        return RequestsResponseInfo(response)


def run(endpoint: EndpointInfo[EndpointResponse]) -> EndpointResponse:
    return RequestsExecutor.run(endpoint)



from ikea_api.abc import Endpoint, SessionInfo, endpoint
from ikea_api.base_ikea_api import BaseIkeaAPI
from ikea_api.error_handlers import (
    handle_401,
    handle_json_decode_error,
    handle_not_success,
)


class Auth(BaseIkeaAPI):
    def _get_session_info(self) -> SessionInfo:
        url = "https://api.ingka.ikea.com/guest/token"
        headers = self._extend_default_headers(
            {
                "Accept": "*/*",
                "Accept-Language": "en-us",
                "X-Client-Id": "e026b58d-dd69-425f-a67f-1e9a5087b87b",
                "X-Client-Secret": "cP0vA4hJ4gD8kO3vX3fP2nE6xT7pT3oH0gC5gX6yB4cY7oR5mB",
            }
        )
        return SessionInfo(base_url=url, headers=headers)

    @endpoint(handlers=[handle_json_decode_error, handle_401, handle_not_success])
    def get_guest_token(self) -> Endpoint[str]:
        response = yield self._RequestInfo(
            "POST", json={"retailUnit": self._const.country}
        )
        return response.json["access_token"]



from __future__ import annotations

from typing import Any, Literal

from ikea_api.abc import Endpoint, SessionInfo, endpoint
from ikea_api.base_ikea_api import BaseIkeaAPI
from ikea_api.error_handlers import handle_json_decode_error

SearchType = Literal["PRODUCT", "CONTENT", "PLANNER", "REFINED_SEARCHES", "ANSWER"]


class Search(BaseIkeaAPI):
    def _get_session_info(self) -> SessionInfo:
        url = f"https://sik.search.blue.cdtapps.com/{self._const.country}/{self._const.language}/search-result-page"
        return SessionInfo(base_url=url, headers=self._extend_default_headers({}))

    @endpoint(handlers=[handle_json_decode_error])
    def search(
        self,
        query: str,
        *,
        limit: int = 24,
        types: list[SearchType] = ["PRODUCT"],
    ) -> Endpoint[dict[str, Any]]:
        params = {
            "autocorrect": "true",
            "subcategories-style": "tree-navigation",
            "types": ",".join(types),
            "q": query,
            "size": limit,
            "c": "sr",  # API client: sr - search results, sb - search bar
            "v": "20210322",  # API version
        }
        response = yield self._RequestInfo("GET", params=params)
        return response.json



from __future__ import annotations

import json
from typing import Any, Dict, List, cast

from ikea_api.abc import ResponseInfo
from ikea_api.exceptions import AuthError, GraphQLError, JSONError, NotSuccessError


def handle_json_decode_error(response: ResponseInfo) -> None:
    try:
        response.json
    except json.JSONDecodeError:
        raise JSONError(response)


def handle_401(response: ResponseInfo) -> None:
    if response.status_code == 401:
        raise AuthError(response)


def handle_not_success(response: ResponseInfo) -> None:
    if not response.is_success:
        raise NotSuccessError(response)


def handle_graphql_error(response: ResponseInfo) -> None:
    if "errors" in response.json:
        raise GraphQLError(response)
    elif isinstance(response.json, list):
        for dict_ in cast(List[Dict[str, Any]], response.json):
            if "errors" in dict_:
                raise GraphQLError(response)



from __future__ import annotations

from dataclasses import dataclass
from functools import lru_cache


@dataclass(frozen=True)
class Constants:
    user_agent: str = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_6) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/14.0.3 Safari/605.1.15"
    base_url: str = "https://www.ikea.com"
    country: str = "ru"
    language: str = "ru"

    @property
    def local_base_url(self) -> str:
        return f"{self.base_url}/{self.country}/{self.language}"


@lru_cache
def get_default_headers(constants: Constants) -> dict[str, str]:
    return {
        "Accept-Encoding": "gzip, deflate, br",
        "Accept-Language": constants.language,
        "Connection": "keep-alive",
        "User-Agent": constants.user_agent,
        "Origin": constants.base_url,
        "Referer": constants.base_url + "/",
    }



from __future__ import annotations

import asyncio
import re
from typing import TYPE_CHECKING, Any, Iterable, Optional, cast

from ikea_api.constants import Constants

if TYPE_CHECKING:
    import httpx


def parse_item_codes(item_codes: list[str] | str) -> list[str]:
    raw_res: list[str] = re.findall(
        r"\d{3}[, .-]{0,2}\d{3}[, .-]{0,2}\d{2}", str(item_codes)
    )
    regex = re.compile(r"[^0-9]")
    # http://stackoverflow.com/questions/480214/how-do-you-remove-duplicates-from-a-list-in-python-whilst-preserving-order
    return list(dict.fromkeys(regex.sub("", i) for i in raw_res))


def _parse_ingka_pagelink_urls(message: str) -> Iterable[str]:
    base_url = "https://ingka.page.link/"
    postfixes = re.findall("ingka.page.link/([0-9A-z]+)", message)
    for postfix in postfixes:
        yield base_url + postfix


def _get_location_headers(responses: Iterable[httpx.Response]) -> list[str]:
    res: list[str] = []
    for response in responses:
        location = cast(Optional[str], response.headers.get("Location"))  # type: ignore
        if location is not None:
            res.append(location)
    return res


async def unshorten_urls_from_ingka_pagelinks(message: str) -> list[str]:
    import httpx

    client = httpx.AsyncClient()
    urls = _parse_ingka_pagelink_urls(message)
    coros = (client.get(url, follow_redirects=False) for url in urls)  # type: ignore
    responses = await asyncio.gather(*coros)
    return _get_location_headers(responses)


def format_item_code(item_code: str) -> str | None:
    matches = parse_item_codes(item_code)
    if not matches:
        return None
    item_code = matches[0]
    return item_code[0:3] + "." + item_code[3:6] + "." + item_code[6:8]


def translate_from_dict(
    constants: Constants, dictionary: dict[str, dict[str, Any]], v: str
) -> str:
    lang_dict = dictionary.get(constants.language)
    if lang_dict is None:
        return v
    return lang_dict.get(v, v)



from ikea_api.constants import Constants as Constants
from ikea_api.endpoints.auth import Auth as Auth
from ikea_api.endpoints.cart import Cart as Cart
from ikea_api.endpoints.ingka_items import IngkaItems as IngkaItems
from ikea_api.endpoints.iows_items import IowsItems as IowsItems
from ikea_api.endpoints.order_capture import OrderCapture as OrderCapture
from ikea_api.endpoints.order_capture import (
    convert_cart_to_checkout_items as convert_cart_to_checkout_items,
)
from ikea_api.endpoints.pip_item import PipItem as PipItem
from ikea_api.endpoints.purchases import Purchases as Purchases
from ikea_api.endpoints.search import Search as Search
from ikea_api.endpoints.stock import Stock as Stock
from ikea_api.exceptions import APIError as APIError
from ikea_api.exceptions import AuthError as AuthError
from ikea_api.exceptions import GraphQLError as GraphQLError
from ikea_api.exceptions import ItemFetchError as ItemFetchError
from ikea_api.exceptions import JSONError as JSONError
from ikea_api.exceptions import NotSuccessError as NotSuccessError
from ikea_api.exceptions import ParsingError as ParsingError
from ikea_api.exceptions import ProcessingError as ProcessingError
from ikea_api.exceptions import WrongItemCodeError as WrongItemCodeError
from ikea_api.executors.httpx import run_async as run_async
from ikea_api.executors.requests import run as run
from ikea_api.utils import format_item_code as format_item_code
from ikea_api.utils import parse_item_codes as parse_item_codes

try:
    import pydantic  # type: ignore
except ImportError:
    pass
else:
    from ikea_api.wrappers.wrappers import add_items_to_cart as add_items_to_cart
    from ikea_api.wrappers.wrappers import (
        get_delivery_services as get_delivery_services,
    )
    from ikea_api.wrappers.wrappers import get_items as get_items
    from ikea_api.wrappers.wrappers import get_purchase_history as get_purchase_history
    from ikea_api.wrappers.wrappers import get_purchase_info as get_purchase_info



from __future__ import annotations

from typing import Any

from ikea_api.abc import Endpoint, SessionInfo, endpoint
from ikea_api.base_ikea_api import BaseIkeaAPI
from ikea_api.error_handlers import handle_graphql_error, handle_json_decode_error


class Stock(BaseIkeaAPI):
    def _get_session_info(self) -> SessionInfo:
        url = f"https://api.ingka.ikea.com/cia/availabilities/ru/{self._const.country}"
        headers = self._extend_default_headers(
            {
                "Accept": "application/json;version=2",
                "Referer": "https://www.ikea.com/",
                "X-Client-Id": "b6c117e5-ae61-4ef5-b4cc-e0b1e37f0631",
            }
        )
        return SessionInfo(base_url=url, headers=headers)

    @endpoint(handlers=[handle_json_decode_error, handle_graphql_error])
    def get_stock(self, item_code: str) -> Endpoint[dict[str, Any]]:
        params = {
            "itemNos": [item_code],
            "expand": "StoresList,Restocks,SalesLocations",
        }
        response = yield self._RequestInfo("GET", params=params)
        return response.json



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['ikea_api',
 'ikea_api.endpoints',
 'ikea_api.executors',
 'ikea_api.wrappers',
 'ikea_api.wrappers.parsers']

package_data = \
{'': ['*']}

extras_require = \
{':python_version < "3.10"': ['typing-extensions>=4.0.0,<5.0.0'],
 'all': ['requests', 'pydantic>=1.8.0', 'httpx>=0.22.0'],
 'httpx': ['httpx>=0.22.0'],
 'requests': ['requests'],
 'wrappers': ['pydantic>=1.8.0']}

setup_kwargs = {
    'name': 'ikea-api',
    'version': '2.0.6',
    'description': "Client for several IKEA's APIs",
    'long_description': 'Client for several IKEA APIs.\n\n[![Test](https://github.com/vrslev/ikea-api-client/actions/workflows/test.yml/badge.svg)](https://github.com/vrslev/ikea-api-client/actions/workflows/test.yml)\n[![Python](https://img.shields.io/pypi/pyversions/ikea_api?label=Python)](https://pypi.org/project/ikea_api)\n[![Downloads](https://img.shields.io/pypi/dm/ikea_api?label=Downloads&color=blueviolet)](https://pypi.org/project/ikea_api)\n\n# Features\n\nWith this library you can access following IKEA\'s APIs:\n\n- Cart,\n- Home Delivery and Collect Services (actually, Order Capture),\n- Items info (3 different services),\n- Purchases (history and order info),\n- Search,\n- Stock.\n\nAlso the package:\n\n- Is **backend agnostic**: choose HTTP library you want (async, too!),\n- Is **fully typed and tested**,\n- Has optional wrappers around some APIs based on Pydantic.\n\n# Installation\n\n```bash\npip install ikea_api\n```\n\n- Use [httpx](https://www.python-httpx.org) — awesome async HTTP library —\xa0as backend:\n\n```bash\npip install "ikea_api[httpx]"\n```\n\n- Use [requests](https://docs.python-requests.org)\xa0as backend:\n\n```bash\npip install "ikea_api[requests]"\n```\n\n- Use wrappers:\n\n```bash\npip install "ikea_api[wrappers]"\n```\n\n- Install everything:\n\n```bash\npip install "ikea_api[all]"\n```\n\n# Usage\n\n_ikea_api_ is unusual API client. It decouples I/O from logic for easier testing and maintenance. As a bonus, you can use literally _any_ HTTP library.\nLet\'s have a look at how to work with ikea_api.\n\n```python\nimport ikea_api\n\n# Constants like country, language, base url\nconstants = ikea_api.Constants(country="us", language="en")\n# Search API\nsearch = ikea_api.Search(constants)\n# Search endpoint with prepared data\nendpoint = search.search("Billy")\n```\n\nAs you can see, nothing happened up to this point. Code suggests that we already should get the result of the search but we don\'t. What happened is `search()` returned a data class that contains information about endpoint that can be interpreted by an endpoint runner. There are two built-in: for [requests](https://docs.python-requests.org) (sync) and [httpx](https://www.python-httpx.org) (async), but you can easily write one yourself.\n\nHere\'s how you would use _requests_ one:\n\n```python\nikea_api.run(endpoint)\n```\n\nAnd _httpx_ one:\n\n```python\nawait ikea_api.run_async(endpoint)\n```\n\n`ikea_api.run_async()` is async function, so you have to "await" it or run using `asyncio.run()`.\n\n## Endpoints reference\n\n### 🔑 Authorization\n\nFirst time you open ikea.com, guest token is being generated and stored in cookies. Same thing must be done in here before using any endpoint.\n\nThis token expires in 30 days.\n\n```python\nikea_api.Auth(constants).get_guest_token()\n```\n\nPreviously you could login as user (with login and password), but now there\'s very advanced telemetry that I wouldn\'t be able to solve in hundred years 🤪\n\n### 🛒 Cart\n\nWith this endpoint you can do everything you can using IKEA\'s frontend:\n\n```python\ncart = ikea_api.Cart(constants, token=...)\n```\n\n- Show the cart\n\n```python\ncart.show()\n```\n\n- Clear it\n\n```python\ncart.clear()\n```\n\n- Add, update and delete items\n\n```python\ncart.add_items({"30457903": 1})  # { item_code: quantity }\n\ncart.update_items({"30457903": 5})\n\ncart.remove_items(["30457903"])\n```\n\n- Set and clear coupon\n\n```python\ncart.set_coupon(...)\n\ncart.clear_coupon()\n```\n\n- and even copy another user\'s cart.\n\n```python\ncart.copy_items(source_user_id=...)\n```\n\nYou can edit your user\'s actual cart if you use authorized token (copy-paste from cookies).\n\n> 💡\xa0There\'s wrapper that clears current cart and adds items with error handling: if requested item doesn\'t exist, the function just skips it and tries again.\n>\n> ```python\n> ikea_api.add_items_to_cart(  # Function returns items that can\'t be added. In this case: [\'11111111\']\n>     cart=cart,\n>     items={\n>         "30457903": 1,\n>         "11111111": 2,  # invalid item that will be skipped\n>     },\n> )\n> ```\n\n### 🚛 Order Capture\n\nCheck pickup or delivery availability. If you need to know whether items are available _in stores_, use [Item availability endpoint](#%F0%9F%9F%A2-item-availability) or [ikea-availability-checker](https://github.com/Ephigenia/ikea-availability-checker).\n\n```python\norder = ikea_api.OrderCapture(constants, token=token)\n\ncart_show = run(cart.show())\nitems = ikea_api.convert_cart_to_checkout_items(cart_show)\n\ncheckout_id = run(order.get_checkout(items))\nservice_area_id = run(\n    order.get_service_area(\n        checkout_id,\n        zip_code="02215",\n        state_code="MA",  # pass State Code only if your country has them\n    )\n)\nhome_services = run(order.get_home_delivery_services(checkout_id, service_area_id))\ncollect_services = run(\n    order.get_collect_delivery_services(checkout_id, service_area_id)\n)\n```\n\n> 💡\xa0You can use wrapper to add items to cart (clearing cart before and handling unknown item errors if they appear) and parse response in nice Pydantic models:\n>\n> ```python\n> services = await ikea_api.get_delivery_services(\n>     constants=constants,\n>     token=...,\n>     items={\n>         "30457903": 1,\n>         "11111111": 2,  # invalid item that will be skipped\n>     },\n>     zip_code="101000",\n> )\n> services.delivery_options  # List of parsed delivery services\n> services.cannot_add  # [\'11111111\']\n> ```\n\n### 📦 Purchases\n\n```python\npurchases = ikea_api.Purchases(constants, token=token)\n```\n\n#### History\n\nThis method requires authentication, so if you don\'t have authorized token, it won\'t work.\n\n```python\npurchases.history()\n\n# Get all purchases:\npurchases.history(take=10000)\n\n# Pagination:\npurchases.history(take=10, skip=1)\n```\n\n> 💡 Get parsed response with the wrapper:\n>\n> ```python\n> ikea_api.get_purchase_history(purchases)  # Returns a list of parsed purchases\n> ```\n\n#### Order info\n\n```python\npurchases.order_info(order_number=..., email=...)\n\n# If you have authorized token, you can drop email:\npurchases.order_info(order_number="111111111")\n\n# The method also has other params but they\'re mostly internal:\npurchases.order_info(\n    order_number=...,\n    email=...,\n    queries=...,  # Queries that will be included in request, combine any of: ["StatusBannerOrder", "CostsOrder", "ProductListOrder"]. By default, all of them are included.\n    # Params below are relevant to ProductListOrder\n    skip_products=...,\n    skip_product_prices=...,\n    take_products=...,\n)\n```\n\n> 💡 Get parsed response with the wrapper:\n>\n> ```python\n> ikea_api.get_purchase_info(  # Returns parsed purchase object. Items are not listed.\n>    purchases=purchases,\n>    order_number=...,\n>    email=...,\n> )\n> ```\n\n### 🪑 Item info\n\nGet item specification by item code (product number or whatever). There are 3 endpoints to do this because you can\'t get all the data about all the items using only one endpoint.\n\n```python\niows_items = ikea_api.IowsItems(constants)\niows_items.get_items(["30457903"])\n\ningka_items = ikea_api.IngkaItems(constants)\ningka_items.get_items(["30457903"])\n\npip_item = ikea_api.PipItem(constants)\npip_item.get_item("30457903")\n```\n\n> 💡 You probably won\'t want to use raw APIs when there\'s convenient "smart" wrapper that combines them all and parses basic info:\n>\n> ```python\n> ikea_api.get_items(["30457903"])\n> ```\n\n### 🟢 Item availability\n\nGet availability by item code (product number or whatever).\n\n```python\nstock = ikea_api.Stock(constants)\nstock.get_stock("30457903")\n```\n\n### 🔎 Search\n\nSearch for products in the product catalog by product name. Optionally also specify a maximum amount of returned search results (defaults to 24) and types of required search results.\n\n```python\nsearch = ikea_api.Search(constants)\nsearch.search("Billy")\n\n# Retrieve 10 search results (default is 24)\nsearch.search("Billy", limit=10)\n\n# Configure search results types\nsearch.search(\n    "Billy",\n    types=...,  # Combine any of: ["PRODUCT", "CONTENT", "PLANNER", "REFINED_SEARCHES", "ANSWER"]\n)\n```\n\n### 🛠 Utilities\n\n#### Parse item codes\n\nParse item codes from string or list.\n\n```python\nassert ikea_api.parse_item_codes("111.111.11") == ["11111111"]\nassert ikea_api.parse_item_codes("11111111, 222.222.22") == ["11111111", "22222222"]\nassert ikea_api.parse_item_codes("111") == []\n```\n\n#### Format item code\n\nParse item code and format it.\n\n```python\nassert ikea_api.format_item_code("11111111") == "111.111.11"\nassert ikea_api.format_item_code("111-111-11") == "111.111.11"\nassert ikea_api.format_item_code("111.111.11") == "111.111.11"\n```\n',
    'author': 'Lev Vereshchagin',
    'author_email': 'mail@vrslev.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/vrslev/ikea-api-client',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from ikea_api.abc import BaseAPI
from ikea_api.constants import Constants, get_default_headers


class BaseIkeaAPI(BaseAPI):
    _const: Constants

    def __init__(self, constants: Constants) -> None:
        self._const = constants
        super().__init__()

    def _extend_default_headers(self, headers: dict[str, str]) -> dict[str, str]:
        res = get_default_headers(constants=self._const).copy()
        res.update(headers)
        return res


class BaseAuthIkeaAPI(BaseIkeaAPI):
    token: str

    def __init__(self, constants: Constants, *, token: str) -> None:
        self.token = token
        super().__init__(constants)

    def _extend_default_headers_with_auth(
        self, headers: dict[str, str]
    ) -> dict[str, str]:
        res = super()._extend_default_headers(headers)
        res["Authorization"] = "Bearer " + self.token
        return res



from __future__ import annotations

from typing import Any, Dict

from ikea_api.abc import Endpoint, RequestInfo, ResponseInfo, SessionInfo, endpoint
from ikea_api.base_ikea_api import BaseIkeaAPI
from ikea_api.exceptions import ItemFetchError, WrongItemCodeError

ItemCodeToComboDict = Dict[str, bool]


def _build_url(items: ItemCodeToComboDict) -> str:
    return ";".join(
        f"{'SPR' if is_combintaion else 'ART'},{item_code}"
        for item_code, is_combintaion in items.items()
    )


class IowsItems(BaseIkeaAPI):
    items: ItemCodeToComboDict

    def _get_session_info(self) -> SessionInfo:
        url = f"https://iows.ikea.com/retail/iows/{self._const.country}/{self._const.language}/catalog/items/"
        headers = self._extend_default_headers(
            {
                "Accept": "application/vnd.ikea.iows+json;version=2.0",
                "Referer": f"{self._const.local_base_url}/shoppinglist/",
                "consumer": "MAMMUT#ShoppingCart",
                "contract": "37249",
            }
        )
        return SessionInfo(base_url=url, headers=headers)

    def _build_request(self) -> RequestInfo:
        return self._RequestInfo("GET", _build_url(self.items))

    def _handle_response(
        self, response: ResponseInfo, relapse: int
    ) -> list[dict[str, Any]] | None:
        if response.status_code == 404 and len(self.items) == 1:
            if relapse != 0:
                raise WrongItemCodeError(response)

            item_code = list(self.items.keys())[0]
            self.items[item_code] = not self.items[item_code]
            return

        if not response.text:
            return

        if "ErrorList" not in response.json or relapse == 2:
            if "RetailItemCommList" in response.json:
                return response.json["RetailItemCommList"]["RetailItemComm"]
            return [response.json["RetailItemComm"]]

        errors: list[dict[str, Any]] | dict[str, Any] = response.json["ErrorList"][
            "Error"
        ]
        if isinstance(errors, dict):
            errors = [errors]

        for err in errors:
            if err.get("ErrorCode", {}).get("$") != 1101:
                # Not error with item type
                raise ItemFetchError(response, err)

            err_list: dict[str, Any] = {
                attr["Name"]["$"]: attr["Value"]["$"]
                for attr in err["ErrorAttributeList"]["ErrorAttribute"]
            }
            item_code = str(err_list["ITEM_NO"])
            if relapse == 0:  # Probably wrong item type: try to switch it
                self.items[item_code] = err_list["ITEM_TYPE"] == "ART"
            else:  # Nope, item is invalid
                del self.items[item_code]

    @endpoint()
    def get_items(self, item_codes: list[str]) -> Endpoint[list[dict[str, Any]]]:
        self.items = {i: False for i in item_codes}

        for relapse in range(3):
            response = yield self._build_request()
            data = self._handle_response(response, relapse)
            if data is not None:
                return data

        raise NotImplementedError
